#include <linux/input.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <QTimer>
#include <QDBusConnection>
#include <QDebug>
#include <QMutex>
#include "sd_bridge.h"
#include "nmInterfaceSD.h"
#include "audio.h"
#include "powermanager.h"
#include "nm_dbus_server.h"

#define SD_SERVICE "org.freedesktop.systemd1"
#define SD_PATH    "/org/freedesktop/systemd1"

const char devinputevent[]="/dev/input/";
#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define LONG(x) ((x)/BITS_PER_LONG)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define test_bit(bit, array)   ((array[LONG(bit)] >> OFF(bit)) & 1)


bool isheadphoneplugged = false;
bool ismicplugged = false;
pm_thread *pthr=NULL;
OrgFreedesktopSystemd1ManagerInterface *ifaced=NULL;
QMutex mp;
extern nm_dbus_server *iface;

pm_thread::pm_thread(int fdi):QThread()
{

	context = NULL;
	mainloop_api = NULL;
	isconnected = false;
	fd = fdi;
	ev_wait = CreateEvent(NULL,true,false,NULL);
	ev_stop_loop_end = CreateEvent(NULL,true,false,NULL);
	SetEvent(ev_stop_loop_end);
	connect_server(1);
}
bool pm_thread::connect_server(int max_try)
{
	int error_id;	

	if (isconnected)
		return true;
	if (max_try <=0)
		max_try = 15;

	tpm_wait(ev_stop_loop_end,error_id,-1);//waiting for event loop thread to cleanup
	for (int i=0;i < max_try;i++){
		m = pa_mainloop_new();
		if (!m)
			return false;
		mainloop_api = pa_mainloop_get_api(m);
		if (!context)
			context = pa_context_new(mainloop_api, "main");
		if (!context)
			return false;
		pa_context_set_state_callback(context, &pm_thread::context_state_callback, NULL);	
		error_id = pa_context_connect(context, NULL, PA_CONTEXT_NOFLAGS , NULL);
		if (error_id < 0){
			pa_context_unref(context);context = NULL;
			pa_mainloop_free(m);m=NULL;
			fprintf(stderr,"can not connect to pa server error_id %d %s\n",error_id,(i == (max_try - 1))?"":"retrying");
			if (i == (max_try - 1))
				return false;
			else
				usleep(500000);
		}
		else {
			SetEvent(ev_wait);
			break;
		}
	}
	isconnected = true;
	return isconnected;
}
void pm_thread::disconnect_server()
{	
	if (context){
		pa_context_disconnect(context);
		pa_context_unref(context);context = NULL;
		ResetEvent(ev_wait);
		ResetEvent(ev_stop_loop_end);
		pa_mainloop_quit(m,0);
	}
	isconnected = false;
}
void pm_thread::run()
{
	int ret;
	
	fprintf(stderr,"starting pa thread ... \n");
	tpm_wait(ev_wait,ret,-1);
	while (1){
		fprintf(stderr,"entering loop \n");
		if (pa_mainloop_run(m, &ret) < 0){
		
		}
		pa_mainloop_free(m);m=NULL;
		SetEvent(ev_stop_loop_end);
		fprintf(stderr,"quit loop \n");
		tpm_wait(ev_wait,ret,-1);		
	}
}
void pm_thread::context_state_callback(pa_context *c, void *userdata)
{
	fprintf(stderr,"pa new state %d \n",pa_context_get_state(c));
	switch (pa_context_get_state(c)) {
	case PA_CONTEXT_READY:
		//fprintf(stderr,"pa new state %d \n",pa_context_get_state(c));
		pa_context_set_subscribe_callback(pthr->context, &pm_thread::context_subscribe_callback, NULL);
		pa_context_subscribe(pthr->context,(pa_subscription_mask_t)
							 (PA_SUBSCRIPTION_MASK_SINK|
							  PA_SUBSCRIPTION_MASK_SOURCE|
							  PA_SUBSCRIPTION_MASK_SINK_INPUT|
							  PA_SUBSCRIPTION_MASK_SOURCE_OUTPUT|
							  PA_SUBSCRIPTION_MASK_MODULE|
							  PA_SUBSCRIPTION_MASK_CLIENT|
							  PA_SUBSCRIPTION_MASK_SAMPLE_CACHE|
							  PA_SUBSCRIPTION_MASK_SERVER|
							  PA_SUBSCRIPTION_MASK_CARD),
							 NULL,
							 NULL);
		pa_context_get_card_info_list(pthr->context, get_card_info_callback, NULL);
		break;
	case PA_CONTEXT_FAILED:
		fprintf(stderr,("Connection failure: %s\n"), pa_strerror(pa_context_errno(c)));
	default:
		break;
		
	}
}
void pm_thread::simple_callback(pa_context *c, int success, void *userdata)
{
	fprintf(stderr,"pa success %d \n",success);
	if (!success)
		fprintf(stderr,"Failure: %s \n", pa_strerror(pa_context_errno(c)));
}
void pm_thread::context_subscribe_callback(pa_context *c, pa_subscription_event_type_t t, uint32_t idx, void *userdata)
{
	//fprintf(stderr,"pa event %d idx %d\n",t,idx);
	if (t == PA_SUBSCRIPTION_EVENT_CARD){//card added ??
		if (pthr->context)
			pa_context_get_card_info_list(pthr->context, get_card_info_callback, NULL);
	}
	else if (t & PA_SUBSCRIPTION_EVENT_REMOVE ){
		if (t & PA_SUBSCRIPTION_EVENT_CARD){//card removed ??
			auto it = std::find_if(pthr->cards.begin(), pthr->cards.end(), [idx] (pa_card c){ return c.idx == (int)idx;});
			if (it != pthr->cards.end()){
				fprintf(stderr,"pa removing card %s\n",(*it).name.toUtf8().data());
				pthr->cards.erase(std::remove_if(pthr->cards.begin(),pthr->cards.end(),[idx](pa_card c){ return (c.idx == (int)idx);}),pthr->cards.end());
			}
		}
	}
}
void pm_thread::get_card_info_callback(pa_context *c, const pa_card_info *i, int is_last, void *userdata)
{
	fprintf(stderr,"pa card info %d  \n",is_last);
	if (!is_last){
		pa_card c;

		auto it = std::find_if(pthr->cards.begin(), pthr->cards.end(), [i] (pa_card c){ return c.name == i->name;}); 
		if (it == pthr->cards.end()){
			c.name = i->name;
			c.idx = i->index;
			pthr->cards.push_back(c);
			fprintf(stderr,"pa name %s idx %d\n",i->name,i->index);
		}
	}
}
void  pm_thread::set_sink_port(const char *port, const char *value, int delay)
{
	if (!delay)
		pa_context_set_sink_port_by_name(pthr->context, port,value, &pm_thread::simple_callback, NULL);
	else
		QTimer::singleShot(delay,NULL, [port,value]() {fprintf(stderr,"%s %s\n",port,value);if(pthr->context) pa_context_set_sink_port_by_name(pthr->context, port,value, &pm_thread::simple_callback, NULL);});
}
void  pm_thread::set_sink_port(bool isplugged, int delay)
{
	set_sink_port("sink.primary_output",isplugged?"output-wired_headset":"output-speaker", delay);
}
void  pm_thread::set_source_port(const char *port, const char *value, int delay)
{
	if (!delay)
		pa_context_set_source_port_by_name(pthr->context, port,value, &pm_thread::simple_callback, NULL);
	else
		QTimer::singleShot(delay,NULL, [port,value]() {fprintf(stderr,"%s %s\n",port,value);if(pthr->context) pa_context_set_source_port_by_name(pthr->context, port,value, &pm_thread::simple_callback, NULL);});
}
void  pm_thread::set_source_port(bool isplugged, int delay)
{
	set_source_port("source.droid",isplugged?"input-wired_headset":"input-default", delay);
}
sd_dbus_iface::sd_dbus_iface(QObject * parent) : QDBusServiceWatcher(parent)
{
	unit_name = "pulseaudio.service";	
}

sd_dbus_iface* sd_dbus_iface::init_iface(QDBusConnection *connection)
{
	
	if (!connection->isConnected()) {	  
		qDebug() << "nm_init_dbus_iface: Cannot connect to session bus " << connection->lastError().message();
		delete connection; connection = NULL;
		return NULL;
	}
	connectioni = connection;
	ifaced = new OrgFreedesktopSystemd1ManagerInterface(SD_SERVICE,SD_PATH,*connection);
	if (!ifaced)
		return NULL; 

	fprintf (stderr, "Succesfully connected to %s and iface %s \n",SD_SERVICE,SD_PATH);
	ifaced->Subscribe();
	register_on_properties();
	
	QObject::connect(ifaced,&OrgFreedesktopSystemd1ManagerInterface::UnitNew,this , &sd_dbus_iface::on_UnitNew);
	QObject::connect(ifaced,&OrgFreedesktopSystemd1ManagerInterface::UnitRemoved,this , &sd_dbus_iface::on_UnitRemoved);
	QObject::connect(ifaced,&OrgFreedesktopSystemd1ManagerInterface::JobNew,this , &sd_dbus_iface::on_JobNew);
	QObject::connect(ifaced,&OrgFreedesktopSystemd1ManagerInterface::JobRemoved,this , &sd_dbus_iface::on_JobRemoved);
	return this;
}
void sd_dbus_iface::on_UnitNew(const QString &in0, const QDBusObjectPath &in1)
{
	//qDebug() << "new " << in0 << in1;
}
void sd_dbus_iface::on_UnitRemoved(const QString &in0, const QDBusObjectPath &in1)
{
	//qDebug() << "removed " << in0 << in1;
}
void sd_dbus_iface::on_JobNew(uint in0, const QDBusObjectPath &in1, const QString &in2)
{
	//qDebug() << "new job " << in0 << in1 << in2;
	if (in2 == unit_name){
		register_on_properties();
	}
}
void sd_dbus_iface::on_JobRemoved(uint in0, const QDBusObjectPath &in1, const QString &in2, const QString &in3)
{
	//qDebug() << "removed job " << in0 << in1 << in2;
}
void sd_dbus_iface::on_propertiesChanged(QString interface,QVariantMap properties,QStringList invalidatedProps)
{
	//qDebug() <<"on_propertiesChanged";	
	//qDebug() << properties;
	auto i = properties.constBegin();
	while (i != properties.constEnd()) {
		if (i.key() == "SubState"){
			//qDebug() << i.key() << i.value();
			//val dead running
			if (i.value() == "running"){
				if (pthr->connect_server()){
					fprintf(stderr,"we are connected to pa \n");
					jack_set_initial_values(pthr->get_fd(),1000);
				}
			}
			else if (i.value() == "dead"){
				pthr->disconnect_server();
			}
		}
		++i;
	}
}
void sd_dbus_iface::register_on_properties()
{
	QDBusObjectPath op = ifaced->GetUnit(unit_name);
	QString path = op.path();
	
	qDebug() << "pulse " << path;
	if (path == ""){
		path = "/org/freedesktop/systemd1/unit/pulseaudio_2eservice";//in case service was off we can not register slot and on connection will not get PropertiesChanged
	}
	connectioni->connect(SD_SERVICE,path, "org.freedesktop.DBus.Properties", "PropertiesChanged",this,SLOT(on_propertiesChanged(QString,QVariantMap,QStringList)));
}
sd_dbus_iface* sd_init_dbus_iface()
{
	QDBusConnection *connection;
	QDBusConnection c = QDBusConnection::connectToBus(QDBusConnection::SystemBus,"systembus");
	sd_dbus_iface *iface=NULL;

	iface = new sd_dbus_iface();
	if (!iface){	
		return NULL;
	}
	connection  =  new QDBusConnection(c);
	if (!connection){		
		return NULL;
	}
	return iface->init_iface(connection);
}
int pa_init(int fd)
{
	pthr = new pm_thread(fd);	
	pthr->start();
	if (!pthr->context)
		return -1;
	return 0;
}
void input_process_event(int fd)
{
	size_t rb;
	int dsize;
	struct input_event ev[64];

	rb = read(fd, ev, sizeof(ev));
	if (rb < (int) sizeof(struct input_event)) {
		fprintf(stderr,"short read \n");
		return;
	}
	dsize = (int) (rb / sizeof(struct input_event));
	for (int i = 0; i < dsize; i++) {
		fprintf(stderr,"got event type %d code %d val %d \n",ev[i].type,ev[i].code,ev[i].value);
		if (ev[i].type == EV_SW){
			if (ev[i].code == SW_HEADPHONE_INSERT){
				pthr->set_sink_port(ev[i].value, 0);
			}
			if (ev[i].code == SW_MICROPHONE_INSERT)
				pthr->set_source_port(ev[i].value, 0);
		}
		else if (ev[i].type == EV_KEY && ev[i].value == 1){//press
			if (ev[i].code == KEY_POWER){
				iface->powerdown_alert(1, "Power button pressed!");
			}
		}
	}
}

int jack_get_ev_status(int fd,bool &isfound)
{
	unsigned long bit[EV_MAX][NBITS(KEY_MAX)];
	unsigned long state[KEY_CNT] = {0};

	
	isheadphoneplugged = ismicplugged = false;
	isfound = false;
	ioctl(fd, EVIOCGBIT(0, EV_MAX), bit[0]);
	for (unsigned int type = 0; type < EV_MAX; type++) {
		if (test_bit(type, bit[0]) && type != EV_REP) {
			if (type != EV_SW)
				continue;
			ioctl(fd, EVIOCGSW(sizeof(state)), state);
			ioctl(fd, EVIOCGBIT(type, KEY_MAX), bit[type]);
			for (int code = 0; code < KEY_MAX; code++){
				if (test_bit(code, bit[type])) {
					int stateval = test_bit(code, state);
					if (code == SW_HEADPHONE_INSERT || code == SW_MICROPHONE_INSERT || code == SW_JACK_PHYSICAL_INSERT){//our guy
						if (code == SW_HEADPHONE_INSERT){
							isheadphoneplugged = stateval;
						}
						if (code == SW_MICROPHONE_INSERT)
							ismicplugged = stateval;
						isfound = true;
					}
				}
			}
		}
	}
	return 0;
}
int power_get_ev_status(int fd,bool &isfound)
{
	unsigned long bit[EV_MAX][NBITS(KEY_MAX)];
	
	isfound = false;
	ioctl(fd, EVIOCGBIT(0, EV_MAX), bit[0]);
	for (unsigned int type = 0; type < EV_MAX; type++) {
		if (test_bit(type, bit[0]) && type != EV_REP) {
			if (type != EV_KEY)
				continue;
			ioctl(fd, EVIOCGBIT(type, KEY_MAX), bit[type]);
			for (int code = 0; code < KEY_MAX; code++){
				if (test_bit(code, bit[type])) {
					if (code == KEY_POWER){//our guy					
						isfound = true;
					}
				}
			}
		}
	}
	return 0;
}
int get_fd(int fd_type,QString &devpath)
{
	struct dirent **namelist=NULL;
	int ndev;
	int fd;

	sd_init_dbus_iface();
	ndev = scandir(devinputevent, &namelist, NULL, NULL);
	if (ndev <= -1)
		return -1;
	for (int i=0; i< ndev;i++){
		char s[1024];
		int  version;
		bool isfound = false;
		
		snprintf(s,sizeof(s)-1,"%s/%s",devinputevent,namelist[i]->d_name);
		//fprintf(stderr,"checking %s \n",s);
		fd = open(s,O_RDONLY);
		if (fd < 0)
			continue;
		if (ioctl(fd, EVIOCGVERSION, &version)) {
			continue;
		}
		if (fd_type == FD_TYPE_JACK)
			jack_get_ev_status(fd,isfound);
		else
			power_get_ev_status(fd,isfound);
		
		if (isfound){
			devpath = s;
			free(namelist);
			return fd;
		}
		close(fd);
	}
	free(namelist);
	return -1;
}
int jack_get_fd()
{
	int  fd;
	QString devpath;
	
	fd = get_fd(FD_TYPE_JACK,devpath);
	if (fd > 0){
		fprintf(stderr,"found jack plug evdev %s %d %d\n",devpath.toUtf8().data(),isheadphoneplugged, ismicplugged);
		if (pa_init(fd)){
		}
		fprintf(stderr,"isconnected pa %d \n", pthr->isconnected);
		if (pthr->isconnected){
			jack_set_initial_values(fd,1000);
		}
	}
	return fd;
}
int power_get_fd()
{
	int  fd;
	QString devpath;
	
	fd = get_fd(FD_TYPE_POWER,devpath);
	if (fd > 0){
		fprintf(stderr,"found power plug evdev %s\n",devpath.toUtf8().data());	
	}
	return fd;
}
void jack_set_initial_values(int fd,int delay)
{
	bool isfound;

	jack_get_ev_status(fd,isfound);
	if (pthr->isconnected){
		pthr->set_sink_port(isheadphoneplugged, delay);
		pthr->set_source_port(ismicplugged, delay);
	}
}
