#include "nm_dbus_server.h"
#include "powermanagerAdaptor.h"
#include <QDBusConnection>
#include "powermanager.h"

QDBusConnection *sconnection=NULL;

#define NM_SERVICE "org.editec.powermanager"
#define NM_PATH    "/powermanager"
extern QObject *root_object;

nm_dbus_server::nm_dbus_server(QObject * parent) : QObject(parent)
{
}

void  nm_dbus_server::set_time(const QString &s)
{	
	int error_id;
	
	qDebug() << "set_time" << s;
	thr->set_time(s.toUtf8().data());
	
}
QStringList nm_dbus_server::get_battery_info()
{
	QList<QString> l;
	battery b[2];
	int charge_status;
	int bat_type = -1;
	int bat_val = 0;
	double temp = 0;
	double voltage;
	double max_bat = 0;
	double min_bat = 0;
	double cur_bat = 0;
	int error_id;
	
	error_id = thr->send_command(CMD_BAT_1_INF);
	error_id = thr->send_command(CMD_BAT_2_INF);
	error_id = thr->send_command(CMD_GET_CHARGE_INF);
	if (!error_id){
		thr->m.lock();
		b[0] = thr->bat[0];
		b[1] = thr->bat[1];
		charge_status = (thr->charge_status == 4)?1:0;
		thr->m.unlock();

		for (int i=0;i < 2;i++){
			if (b[i].bat_type != BAT_DISCONNECTED){
				bat_type = b[i].bat_type;
				temp = b[i].temp;
				voltage = b[i].voltage;
				min_bat += thr->get_min_battery_voltage();
				max_bat += thr->get_max_battery_voltage();
				cur_bat += voltage;
			}
		}		
		bat_val = (int)100*((cur_bat - min_bat)/(max_bat - min_bat));
		l.push_back(QString::number(bat_type));
		l.push_back(QString::number(temp));
		l.push_back(QString::number(voltage));
		l.push_back(QString::number(charge_status));
		l.push_back(QString::number(bat_val));
		//for android gui need more testing without battery(android reboots)
		{
			int onlinei=(bat_type != BAT_DISCONNECTED)?1:1;//force
			int bat_vali=(bat_type != BAT_DISCONNECTED)?bat_val:99;//force
			
			thr->sysfs_set_value("/sys/devices/platform/soc/200f000.qcom,spmi/spmi-0/spmi0-02/200f000.qcom,spmi:qcom,pmi632@2:qcom,qpnp-smb5/power_supply/usb/online", onlinei);
			thr->sysfs_set_value("/sys/devices/platform/soc/200f000.qcom,spmi/spmi-0/spmi0-02/200f000.qcom,spmi:qcom,pmi632@2:qcom,qpnp-smb5/power_supply/battery/capacity",bat_vali);
		}
	}
	return l;
}
int  nm_dbus_server::get_battery_temp()
{
	
	return 0;
}
QStringList nm_dbus_server::get_app_version()
{
	QList<QString> l;	
	int version[4];
	int error_id;
	
	error_id = thr->send_command(CMD_VERSION_CM);
	if (!error_id){
		thr->m.lock();
		memcpy(&version,thr->version,sizeof(thr->version));
		thr->m.unlock();

		for (int i=0;i < (int)(sizeof(version)/sizeof(int));i++)
			l.push_back(QString::number(version[i]));
	}
	return l;
}
QStringList   nm_dbus_server::get_small_battery_info()
{
	return QStringList();
}
void  nm_dbus_server::update_firmware(const QString &path)
{
	
	QTimer::singleShot(0,NULL, [path]() {thr->send_firmare(path);});
}
nm_dbus_server* nm_dbus_server_init()
{
	QDBusConnection *sconnection;
	PowermanagerAdaptor *nmA;
	nm_dbus_server *iface = NULL;

	Q_UNUSED(nmA);
	sconnection = new QDBusConnection(QDBusConnection::systemBus());
	if (!sconnection->isConnected())
		return NULL;	
	if (!sconnection->isConnected()) {	  
		qDebug() << "nm_dbus_server: Cannot connect to session bus " << sconnection->lastError().message();
		delete sconnection; sconnection = NULL;
		return NULL;
	}
	if (!sconnection->interface()->isServiceRegistered(NM_SERVICE)){
		iface = new nm_dbus_server(NULL);
		nmA = new PowermanagerAdaptor(iface);
		if (!sconnection->registerService(NM_SERVICE))
		{
			qFatal("Could not register service!");
			delete nmA;delete iface;
			return NULL;
		}
		if (!sconnection->registerObject(NM_PATH, iface))
		{
			qFatal("Could not register air object! %s",sconnection->lastError().message().toUtf8().data());
			delete nmA;delete iface;
			return NULL;
		}
	}
	qDebug() << "org.editec.powermanager dbus service is registered";
	return iface;
}
