#ifndef POWERMANAGER_H
#define POWERMANAGER_H
#include "qobjectdefs.h"
#include <QThread>
#include <QMutex>
#include <QTimer>
#include <sys/ioctl.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <poll.h>
#include <sys/types.h>
#include <stdint.h>

#define X_SOH 0x01
#define X_STX 0x02
#define X_ACK 0x06
#define X_NAK 0x15
#define X_EOF 0x04
#define X_CAN 0x18
#define X_CRC 0x43

enum proto_cmds {
	CMD_VERSION_CM 	=	    0x00,
	CMD_VERSION_CC	=	    0x01,
	CMD_CHARGEUR 	=	    0x10,
	CMD_BAT_1_INF   =	    0x20,
	CMD_BAT_1_CYCLE_INF	 =  0x21,
	CMD_BAT_1_ERR_INF =     0x22,
	CMD_BAT_2_INF =		    0x30,
	CMD_BAT_2_CYCLE_INF =   0x31,
	CMD_BAT_2_ERR_INF	=   0x32,
	CMD_SET_TIME 	=	    0x40,
	CMD_GET_TIME 	=	    0x41,
	CMD_PILE 		=       0x42,
	CMD_GET_CHARGE_INF	=   0x50,
	CMD_GET_INTRUSION_INF=	0x60,
	CMD_SET_AlIM	=	    0x70,
	CMD_GET_ALIM_INF =	    0x90,	
	CMD_SOFT_REBOOT =		0xC0,
	CMD_SOFT_DOWNLOAD =	    0xC1,
	CMD_SOFT_JUMP =		    0xC3,
	CMD_SOFT_JUMP_NO_CRC =	0xC5,
	CMD_SOFT_DOWNLOAD_REBOOT  = 0xA0,
	CMD_SOFT_DOWNLOAD_REBOOT2 = 0xA3,
};
enum bat_type {
	BAT_DISCONNECTED = 0,
	BAT_LI_LION = 1,
	BAT_LI_FER = 2,
	BAT_CONFLIT = 3,
	BAT_NOT_RECOGNISED = 4
};

#define min(a, b)       ((a) < (b) ? (a) : (b))

struct xmodem_chunk {
        uint8_t start;
        uint8_t block;
        uint8_t block_neg;
        uint8_t payload[1024];
        uint16_t crc;
} __attribute__((packed));
struct xmodem_chunk_sh {
        uint8_t start;
        uint8_t block;
        uint8_t block_neg;
        uint8_t payload[128];
        uint16_t crc;
} __attribute__((packed));

#define CRC_POLY 0x1021

class battery
{
public:
	battery(){bat_type = BAT_DISCONNECTED; temp = 0; voltage = 0;}
	
	battery(const battery & i)
	{
		bat_type = i.bat_type;
		temp = i.temp;
		voltage = i.voltage;
	}
	
	int bat_type;
	double temp;
	double voltage;
};
class job_modem_thread : public QThread
{
	Q_OBJECT
public:
	job_modem_thread();
	void run();
	bool is_handle_sig(pollfd *h, int result, int i);
	int  open_port(const char *port,bool normal);
	int  send_command(int cmd,bool iswait=true);	
	int  set_time(const char *s);
	int  set_power(bool ison);
	int  xymodem_send(const char *filename, int protocol,int *crc32, int wait);
	int  send_firmare(QString filename);
	int  get_boot_mode(){return boot_mode;};
	void sysfs_set_value(const char *path,int value);
	double  get_min_battery_voltage();
	double  get_min_battery_voltage_crytical();
	double  get_max_battery_voltage();
	bool    get_battery_plugged_llt5();
	bool    get_battery_charging_llt5();
	double  get_battery_voltage_llt5();
	battery bat[2];
	int  charge_status;
	int version[4];
	QMutex m;
	QTimer timersht;
	QTimer timerpoff;
private:
	int fd;
	int fdj;
	int fdp;
	int g_write_delay;
	int ev_wait;
	int ev_terminate;
	bool issettime;
	int  boot_mode;
	typedef struct 
	{
		unsigned char adresse;
		unsigned char data[8];
		unsigned int  crc;
		unsigned char fin_de_ligne;
	}Trame;
	Trame DataReceived();
	void Get_trame(Trame trame);
	void Send_trame(Trame trame);	
	int send_data(char *b, int b_len);
	unsigned int calc_crc32 (char *data, int len);
	unsigned int Calcul_crc(Trame trame);
	int	base_Hex(int c);	
	const char *u32_to_str(unsigned int val);
	bool  run_process(const QString &pname);
	void  gpio_get_value_by_fd(int fd,int *value);
private slots:
	void timersht_update();
	void timerpoff_update();
};
int CreateEvent(char *p1,bool p2,bool p3,char *p4);
bool SetEvent2(int ev,u_int64_t u);
bool SetEvent(int ev);
bool ResetEvent(int ev);
void tpm_wait(int fd,int &b,int t);
extern job_modem_thread *thr;

#endif
